/*
	Copyright © 2003 RiskCare Ltd. All rights reserved.
	Copyright © 2010 SvgNet & SvgGdi Bridge Project. All rights reserved.
	Copyright © 2015 Rafael Teixeira, Mojmír Němeček, Benjamin Peterson and Other Contributors

	Original source code licensed with BSD-2-Clause spirit, treat it thus, see accompanied LICENSE for more
*/
using System;
using System.Collections;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Globalization;
using System.Text.RegularExpressions;
using System.Drawing.Imaging;

namespace SvgNet.SvgTypes
{
	public class SvgPen : ICloneable
	{
		public SvgBrush Brush { get; set; }
		public SvgColor Color { get { return new SvgColor(Brush); } }
		public float Width { get; set; }

		public SvgPen(SvgBrush b, float w = 1.0f, 
			DashStyle style = System.Drawing.Drawing2D.DashStyle.Solid,
			DashCap cap = System.Drawing.Drawing2D.DashCap.Flat,
			float offset = 0, float[] pattern = null,
			PenAlignment align = System.Drawing.Drawing2D.PenAlignment.Center,
			LineJoin join = System.Drawing.Drawing2D.LineJoin.Miter, 
			LineCap scap = System.Drawing.Drawing2D.LineCap.Flat,
			LineCap ecap = System.Drawing.Drawing2D.LineCap.Flat,
			CustomLineCap cscap = null,
			CustomLineCap cecap = null,
			Matrix at = null) 
		{  
			Brush = b;
			Width = w;
			Alignment = align;
			DashCap = cap;
			DashPattern = pattern;
			DashStyle = style;
			LineJoin = join;
			Transform = at;
			StartCap = scap;
			EndCap = ecap;
			CustomStartCap = cscap;
			CustomEndCap = cecap;
		}

		public SvgPen(Pen pen) {
			Brush = new SvgBrush(((SolidBrush)pen.Brush).Color);
			Width = pen.Width;
			Alignment = pen.Alignment;
			DashCap = pen.DashCap;
			DashPattern = pen.DashPattern;
			DashStyle = pen.DashStyle;
			LineJoin = pen.LineJoin;
			Transform = pen.Transform;
			StartCap = pen.StartCap;
			EndCap = pen.EndCap;
			try {
				CustomStartCap = pen.CustomStartCap;
				CustomEndCap = pen.CustomEndCap;
			}
			catch {
			}
		}
		
		public virtual object Clone() {
			return new SvgPen((SvgBrush)Brush.Clone(), Width, DashStyle,
				DashCap, DashOffset, DashPattern, Alignment,
				LineJoin, StartCap, EndCap, 
				CustomStartCap, CustomEndCap,
				(Matrix)Transform.Clone() );
		}
 
		public Pen Pen { 
			get { 
				Pen pen = new Pen(Brush.Brush, Width); 
    			pen.Alignment = Alignment;
    			pen.DashCap = DashCap;
    			pen.DashPattern = DashPattern;
    			pen.DashStyle = DashStyle;
    			pen.LineJoin = LineJoin;
    			pen.Transform = Transform;
    			pen.StartCap = StartCap;
    			pen.EndCap = EndCap;
				pen.CustomEndCap = CustomEndCap;
				pen.CustomStartCap = CustomStartCap;
				return pen;
			} 
		}

		public PenAlignment Alignment { get; set; }
		public DashCap DashCap { get; set; }
		public float DashOffset { get; set; }
		public float[] DashPattern { get; set; }
		public DashStyle DashStyle { get; set; }
		public LineJoin LineJoin { get; set; }
		public float MiterLimit { get; set; }
		public PenType PenType { get { return System.Drawing.Drawing2D.PenType.SolidColor; } }
		public LineCap EndCap { get; set; }
		public LineCap StartCap { get; set; }
		public Matrix Transform { get; set; }
		public CustomLineCap CustomEndCap { get; set; }
		public CustomLineCap CustomStartCap { get; set; }
	}
	
	public class SvgBrush : ICloneable 
	{
		public int R { get; set; }
		public int G { get; set; }
		public int B { get; set; }
		public int A { get; set; }

		public float C { get; set; }
		public float M { get; set; }
		public float Y { get; set; }
		public float K { get; set; }

		public bool HasCmyk { get; set; }

		public SvgBrush (int r, int g, int b, int a = 255) {
			R = r;
			G = g;
			B = b;
			A = a;

			C = M = Y = K = 0;
			HasCmyk = false;

			checkIfWhiteOrBlack();
		}

		public SvgBrush (
			float c, float m, float y, float k,
			int r, int g, int b, int a = 255) 
		{
			C = c;
			M = m;
			Y = y;
			K = k;

			R = r;
			G = g;
			B = b;
			A = a;

			HasCmyk = true;
		}
		
		private const int K_EOF = 0;
		private const int K_LBR = 1;
		private const int K_RBR = 2;
		private const int K_NUM = 3;
		private const int K_RGB = 4;
		private const int K_COM = 5;
		private const int K_CMY = 6;
		private const int K_UNK = 7;
		private const int K_HNM = 8;
		private const int K_FLN = 9;
		private const int K_ICC = 10;

		private static string ackTok(string s, ref int j) {
			int k = j;
			while(j < s.Length && (s[j] == '-' || Char.IsLetterOrDigit(s[j]) )) ++j;
			string z = s.Substring(k, j - k);
			return z;
		}

		private static void skipSpaces(string s, ref int j) {
			while(j < s.Length && (Char.IsWhiteSpace(s[j]) || s[j] == '-')) ++j;
		}

		private static string nextToken(string s, ref int j, out int kind) {
			if(j >= s.Length) {
				kind = K_EOF;
				return "";
			}
			
			skipSpaces(s, ref j);
			if(s[j] == '#' || 
					(s.Length > j + 2 && s[j] == '0' && s[j+1] == 'x')) { 
				// hex num
				kind = K_HNM;
				j += s[j] == '#' ? 1 : 2;
				return ackTok(s, ref j);
			}
			else if(Char.IsDigit(s[j])) { 
				// just number
				kind = K_NUM;
				string z0 = ackTok(s, ref j);
				if(s[j] == '.') {
					++j;
					string z1 = ackTok(s, ref j);
					kind = K_FLN;
					z0 += '.' + z1;
				}
				return z0;
			}
			else if(s[j] == ',' || s[j] == ';') {
				kind = K_COM;
				++j;
				return ",";
			}
			else if(s[j] == 'r' || s[j] == 'R') {
				kind = K_RGB;
				return ackTok(s, ref j);
			}
			else if(s[j] == 'i' || s[j] == 'I') { // icc-color
				kind = K_ICC;
				return ackTok(s, ref j);
			}
			else if(s[j] == 'c' || s[j] == 'C') {
				kind = K_CMY;
				return ackTok(s, ref j);
			}
			else if(s[j] == '(' || s[j] == '[') {
				kind = K_LBR;
				++j;
				return "(";
			}
			else if(s[j] == ')' || s[j] == ']') {
				kind = K_RBR;
				++j;
				return "(";
			}
			else {
				kind = K_UNK;
				return ackTok(s, ref j);
			}
		}

		public SvgBrush (String colorString) {
			// possible forms are
			// [#|0x]xxxxxxxx | rgb(x,y,z) | rgba (x,y,z,a) [, cmyk(x,y,z,a)]
			// black | white

			if (string.IsNullOrEmpty(colorString))
			{
				throw new ArgumentException("Color value cannot be empty", "colorString");
			}

			int k;
			int j = 0;
			long argb, cp;
			float[] f = new float[4];
			int t;
			for(t = 0; t < 4; ++t) f[t] = 0.0f;
			HasCmyk = false;

 			string tok = nextToken(colorString, ref j, out k);

 			if (k == K_EOF) // empty string
 				throw new ArgumentException(colorString + ": unexpected EOF");

 			// named colors
 			if(colorString.ToLower().Trim().Equals("white") || colorString.ToLower().Trim().Equals("control")) {
				C = White.C;
				M = White.M;
				Y = White.Y;
				K = White.K;

				R = White.R;
				G = White.G;
				B = White.B;
				A = White.A;

				HasCmyk = true;
				//Console.WriteLine("SvgBrush('"+colorString+"') => " + this);
 				return;	
 			}

 			if(colorString.ToLower().Trim().Equals("black")) {
				C = Black.C;
				M = Black.M;
				Y = Black.Y;
				K = Black.K;

				R = Black.R;
				G = Black.G;
				B = Black.B;
				A = Black.A;

				HasCmyk = true;
				//Console.WriteLine("SvgBrush('"+colorString+"') => " + this);
 				return;	
 			}

 			switch(k) {
 			case K_HNM: // classic ARGB
 				argb = Int32.Parse(tok, NumberStyles.HexNumber) | unchecked((int)0xFF000000);
 				B = (int)( argb        & 0xFF);
 				G = (int)((argb >>  8) & 0xFF);
 				R = (int)((argb >> 16) & 0xFF);
 				A = (int)((argb >> 24) & 0xFF);
				checkIfWhiteOrBlack();
 				break;
 			case K_RGB:  // Note: it composes to ABGR not ARGB
 				argb = 0;
 				tok = nextToken(colorString, ref j, out k);
 				if(k != K_LBR)
 					throw new ArgumentException(
 						colorString + ": unexpected token `" + tok + "` before pos " + j + 
 						", expected `(` after rgb/rgba");
 				for(t = 0; t < 4;) {
	 				tok = nextToken(colorString, ref j, out k);
	 				if(k == K_RBR) break;
	 				if(k == K_COM) continue;
	 				if(k == K_EOF)
		 				throw new ArgumentException(
 							colorString + ": unexpected EOF within rgb(...) specification, `)` expected");

	 				switch(k) {
	 				case K_NUM:
	 					cp = Int32.Parse(tok);
	 					break;
	 				case K_HNM:
	 					cp = Int32.Parse(tok, NumberStyles.HexNumber);
	 					break;
	 				case K_FLN:
	 					cp = (int)(Single.Parse(tok, CultureInfo.InvariantCulture) * 255.0);
	 					break;
	 				default:
		 				throw new ArgumentException(
 							colorString + ": unexpected token `" + tok + "` before pos " + j + 
		 					"within RGB spec, expected comma-separated list of numbers");
	 				}
	 				if(t > 0) cp <<= 8 * t;
	 				argb |= cp;
	 				++t;
 				}

 				if(t < 4) argb |= 0xFF000000;
 				if(k != K_RBR) 
 					tok = nextToken(colorString, ref j, out k);
 				if(k != K_RBR)
 					throw new ArgumentException(
 						colorString + ": unexpected token `" + tok + "` before pos " + j + 
 						", expected `)` after rgb/rgba");
 			
 			    // abgr
 				R = (int)( argb        & 0xFF);
 				G = (int)((argb >>  8) & 0xFF);
 				B = (int)((argb >> 16) & 0xFF);
 				A = (int)((argb >> 24) & 0xFF);
				checkIfWhiteOrBlack();
 				break;
 			case K_UNK:
 			default:
 				throw new ArgumentException(
 					colorString + ": unexpected token `" + tok + "` before pos " + j + 
 					", expected one of rgb(...), rgba (...), #RRGGBB, 0xRRGGBB");
 			}

			tok = nextToken(colorString, ref j, out k);
			if(k == K_EOF){ 
				checkIfWhiteOrBlack();
				//Console.WriteLine("SvgBrush('"+colorString+"') => " + this);
				return;
			}
			if(k == K_COM) 
				tok = nextToken(colorString, ref j, out k);
			if(k != K_CMY && k != K_ICC)
 				throw new ArgumentException(
 					colorString + ": unexpected token `" + tok + "` before pos " + j + 
 					", expected cmyk(...) | icc-color(CMYK,...) specification");
			tok = nextToken(colorString, ref j, out k);
			if(k != K_LBR)
				throw new ArgumentException(
					colorString + ": unexpected token `" + tok + "` before pos " + j + 
					", expected `(` after cmyk");

			for(t = 0; t < 4;) {
 				tok = nextToken(colorString, ref j, out k);
 				if(k == K_RBR)
	 				throw new ArgumentException(
						colorString + ": unexpected ')' within cmyk(...) specification, " + t + " of 4 parts are found");
 				if(k == K_COM) continue;
 				if(k == K_CMY) continue; // icc-color(CMYK,...
 				if(k == K_EOF)
	 				throw new ArgumentException(
						colorString + ": unexpected EOF within cmyk(...) specification, `)` expected");

 				switch(k) {
 				case K_NUM:
					f[t] = Single.Parse(tok, CultureInfo.InvariantCulture);
 					if(f[t] > 1) f[t] /= 255.0f;
 					break;
 				case K_HNM:
 					f[t] = Int32.Parse(tok, NumberStyles.HexNumber) / 255.0f;
 					break;
 				case K_FLN:
					f[t] = Single.Parse(tok, CultureInfo.InvariantCulture);
 					break;
 				default:
	 				throw new ArgumentException(
						colorString + ": unexpected token `" + tok + "` before pos " + j + 
	 					"within cmyk spec, expected comma-separated list of numbers");
 				}
 				++t;
			}

			tok = nextToken(colorString, ref j, out k);
			if(k == K_COM) 
				tok = nextToken(colorString, ref j, out k);
			if(k != K_RBR)
				throw new ArgumentException(
					colorString + ": unexpected token `" + tok + "` before pos " + j + 
					", expected `)` after cmyk(...");
			tok = nextToken(colorString, ref j, out k);
			if(k == K_COM) 
				tok = nextToken(colorString, ref j, out k);
			if(k != K_EOF)
				throw new ArgumentException(
					colorString + ": unexpected token `" + tok + "` before pos " + j + 
					", expected EOF");
			C = f[0];
			M = f[1];
			Y = f[2];
			K = f[3];
			HasCmyk = true;
			A = 255;
			//Console.WriteLine("SvgBrush('"+colorString+"') => " + this);
		}

		private void checkIfWhiteOrBlack() {
			if(this.Equals(White) && !HasCmyk) { // Alpha is not used in comparisons
				C = White.C;
				M = White.M;
				Y = White.Y;
				K = White.K;
				HasCmyk = true;
			}
			else if(this.Equals(Black) && !HasCmyk) {
				C = Black.C;
				M = Black.M;
				Y = Black.Y;
				K = Black.K;
				HasCmyk = true;
			}
		}

		public static readonly SvgBrush Black = new SvgBrush(0.0f, 0.0f, 0.0f, 1.0f, 0, 0, 0);
		public static readonly SvgBrush White = new SvgBrush(0.0f, 0.0f, 0.0f, 0.0f, 255, 255, 255);
		public static readonly SvgBrush None  = new SvgBrush(128, 0, 0, 1); // indesign uses it!

		public SvgBrush (SolidBrush b) : this(b.Color.R, b.Color.G, b.Color.B, b.Color.A) {}
		public SvgBrush (Color b) : this(b.R, b.G, b.B, b.A) {}

		public virtual object Clone() {
			return HasCmyk ? new SvgBrush(C, M, Y, K, R, G, B, A) : new SvgBrush(R, G, B, A);
		}

		private String format(float f) {
			return String.Format(CultureInfo.InvariantCulture, "{0:0.0###}", f);
		}

		public override string ToString () {
			if(this.Equals(Black)) return "black";
			if(this.Equals(None)) return "none";

			string s = "rgb(";
			s += R.ToString(CultureInfo.InvariantCulture);
			s += ",";
			s += G.ToString(CultureInfo.InvariantCulture);
			s += ",";
			s += B.ToString(CultureInfo.InvariantCulture);
			s += ")";

			if(false && HasCmyk) { /* VK: disable icc for browsers */
				s += " icc-color(CMYK,";
				s += format(C);
				s += ",";
				s += format(M);
				s += ",";
				s += format(Y);
				s += ",";
				s += format(K);
				s += ")";
			}
			return s;
		}

		public override int GetHashCode() {
			return (int)(R << 16 + G << 8 + B);
		}

		public override bool Equals(object o) {
			if(o == null) return false;
			if(o is SvgPen) o = ((SvgPen)o).Brush;
			if(o is SvgColor) o = ((SvgColor)o).Brush;
			if(!(o is SvgBrush)) return false;
			SvgBrush b = (SvgBrush)o;
			return R == b.R && G == b.G && B == b.B;
		}

		public Brush Brush { get { return new SolidBrush(System.Drawing.Color.FromArgb(A, R, G, B)); } }
		public SvgColor Color { get { return new SvgColor(this); } }
	}

	public class SvgColor : ICloneable
	{
		private SvgBrush _col;
		private string _original_string;

		private static Hashtable _stdcols = new Hashtable();

		public int A { get { return _col.A; } }

		public SvgColor(SvgBrush c) {
	    	_col = c;
		}

		public SvgColor(string s) {
	    	_col = new SvgBrush(s);
			_original_string = s;
		}

		public SvgColor(SvgBrush c, string s) {
	    	_col = c;
			_original_string = s;
		}

		public SvgColor(Brush c) {
	    	_col = new SvgBrush((SolidBrush)c);
		}

		public SvgColor(Brush c, string s) {
	    	_col = new SvgBrush((SolidBrush)c);
			_original_string = s;
		}

		public SvgBrush Brush {
			get{return _col;}
			set{_col = value;}
		}

		public System.Drawing.Color BrushColor { get { return System.Drawing.Color.FromArgb(_col.A, _col.R, _col.G, _col.B); } }

		public static SvgColor FromName(String name) { return new SvgColor( name );	}
		public virtual object Clone() {	return new SvgColor((SvgBrush)_col.Clone(), _original_string); }

		public override string ToString()
		{
			return _col.ToString();
		}

		public static implicit operator SvgColor(SvgBrush c)
		{
			return new SvgColor(c);
		}

		public static implicit operator SvgColor(string s)
		{
			return new SvgColor(s);
		}
	}
}
